function RaycastWeaponImage::onMount(%this, %obj, %slot)
{
  WeaponImage::onMount(%this, %obj, %slot);
}

function RaycastWeaponImage::onUnMount(%this, %obj, %slot)
{
  WeaponImage::onUnMount(%this, %obj, %slot);
}

function RaycastWeaponImage::onFire(%this, %obj, %slot)
{
  if (%this.range <= 0)
  {
    return;
  }

  %obj.hasShotOnce = 1;

  if (getSimTime() - %obj.lastFireTime < %this.minShotTime)
  {
    return;
  }

  %obj.lastFireTime = getSimTime();
  %client = %obj.client;

  if (isObject(%client.miniGame) && getSimTime() - %client.lastF8Time < 3000)
  {
    return;
  }

  if (%this.shotMask $= "")
  {
    %mask =
      $TypeMasks::PlayerObjectType |
      $TypeMasks::VehicleObjectType |
      $TypeMasks::TerrainObjectType |
      $TypeMasks::FxBrickObjectType |
      $TypeMasks::StaticShapeObjectType;
  }
  else
  {
    %mask = %this.shotMask;
  }

  if (%this.melee)
  {
    %origin = %obj.getEyePoint();
    %vector = %obj.getEyeVector();
    // %vector = %obj.getMeleeVector();
  }
  else
  {
    %origin = %obj.getMuzzlePoint(0);

    // Get the player's LOS point
    %eyePoint = %obj.getEyePoint();
    %eyeVector = %obj.getEyeVector();

    %eyeEnd = vectorAdd(%eyePoint, vectorScale(%eyeVector, %this.range));
    %eyeRay = containerRayCast(%eyePoint, %eyeEnd, %mask, %obj);

    if (%eyeRay)
    {
      %eyeEnd = getWords(%eyeRay, 1, 3);
    }

    // Aim a vector at it
    %vector = vectorNormalize(vectorSub(%eyeEnd, %origin));
  }

  %count = 1;

  if (%this.shotCount !$= "")
  {
    %count = %this.shotCount;
  }

  for (%i = 0; %i < %count; %i++)
  {
    %spread = %this.getSpread(%obj, %slot, %i);

    if (%spread $= "0 0 0")
    {
      %direction = %vector;
    }
    else
    {
      %matrix = matrixCreateFromEuler(%spread);
      %direction = matrixMulVector(%matrix, %vector);
    }

    %rayVector = vectorScale(%direction, %this.range);
    %rayEnd = vectorAdd(%origin, %rayVector);

    // This is where the magic happens.
    %ray = containerRayCast(%origin, %rayEnd, %mask, %obj);

    if (%ray)
    {
      %rayEnd = getWords(%ray, 1, 3);
      %this.onHitObject(%obj, getWord(%ray, 0), %rayEnd, getWords(%ray, 4, 6), %rayVector);
    }
    else
    {
      %this.onMiss(%obj, %rayEnd, %rayVector);
    }

    if (isObject(%this.tracer))
    {
      // ...
    }
  }
}

function getRandomScalar()
{
  return getRandom() * 2 - 1;
}

function Player::getMeleeVector(%this)
{
  %forward = %this.getForwardVector();
  %eye = %this.getEyeVector();

  %xEye2 = mPow(getWord(%eye, 0), 2);
  %yEye2 = mPow(getWord(%eye, 1), 2);

  %yaw = mATan(getWord(%forward, 0), getWord(%forward, 1));
  %pitch = mATan(getWord(%eye, 2), mSqrt(%xEye2 + %yEye2));

  %length = mCos(%yaw);

  %x = %length * mCos(%pitch);
  %y = mSin(%yaw);
  %z = %length * mSin(%pitch);

  return %x SPC %y SPC %z;
}

function RaycastWeaponImage::getSpread(%this, %obj, %slot, %i)
{
  %limit = %obj.getDataBlock().maxForwardSpeed;
  %moving = %this.spreadMoving $= "" ? %this.spread : %this.spreadMoving;

  %spread = lerp(%this.spread, %moving, vectorLen(%obj.getVelocity()) / %limit);

  if (%this.useSourceSpread)
  {
    %spread = mATan(%spread, 1);
  }
  else
  {
    %spread = mDegToRad(%spread);
  }

  if (%spread > 0)
  {
    %scalars = getRandomScalar() SPC getRandomScalar() SPC getRandomScalar();
    return vectorScale(%scalars, %spread);
  }

  return "0 0 0";
}

function RaycastWeaponImage::onHitObject(%this, %obj, %col, %pos, %normal, %vector)
{
  if (isObject(%this.playerExplosion) && (%col.getType() & $TypeMasks::PlayerObjectType))
  {
    %explosion = %this.playerExplosion;
  }
  else
  {
    %explosion = %this.explosion;
  }

  if (isObject(%explosion))
  {
    %explosion = new Projectile()
    {
      datablock = %explosion;

      initialPosition = %pos;
      initialVelocity = 0;
    };

    MissionCleanup.add(%explosion);

    %explosion.setScale(%obj.getScale());
    %explosion.explode();
  }

  if (%col.getType() & ($TypeMasks::PlayerObjectType | $TypeMasks::VehicleObjectType))
  {
    if (isObject(%col.spawnBrick) && %col.spawnBrick.getGroup().client == %obj.client)
    {
      %spawned = 1;
    }

    if (miniGameCanDamage(%obj, %col) == 1 || %spawned)
    {
      %this.damage(%obj, %col, %pos, %normal, %vector);
    }
  }
}

function RaycastWeaponImage::onMiss(%this, %obj, %pos, %vector)
{
}

function RaycastWeaponImage::damage(%this, %obj, %col, %pos, %normal, %vector)
{
  if (%this.directDamage <= 0)
  {
    return;
  }

  %type = $DamageType::Direct;

  if (%this.directDamageType)
  {
    %type = %this.directDamageType;
  }

  %col.damage(%obj, %pos, %this.directDamage, %type);

  if (%this.impactImpulse)
  {
    %col.applyImpulse(%pos, vectorScale(%vector, %this.impactImpulse));
  }

  if (%this.verticalImpulse)
  {
    %col.applyImpulse(%pos, vectorScale(%col.getUpVector(), %this.verticalImpulse));
  }
}

function lerp(%a, %b, %t)
{
  return %a + (%b - %a) * %t;
}